# encoding: utf-8
#
# Compile and minimizing assets (Javascript/CSS) files

require 'pp'
require 'term/ansicolor'
# mixin String class for term-color methods
class String; include Term::ANSIColor end

require 'yaml'
require 'fileutils'

require 'sass'
require 'closure-compiler'

# configure / compile java script and css assets files
# rubocop:disable MethodLength, ClassLength, CyclomaticComplexity
class LaAssets
  # initialize compiler and setup options
  def initialize(def_file, verbose = false, cleanup = false)
    # default configuration (directory relative to current directory):
    @settings = {
      'public_dir' => 'public',
      'assets_dir' => 'public/assets',
      'js_output_dir' => 'public/javastripts',
      'css_output_dir' => 'public/stylesheets',
      'output_file' => 'config/assets.yml',
      'js_url_prefix' => '/javastripts/',
      'css_url_prefix' => '/stylesheets/',
      'development_prefix' => '/'
    }

    @verbose = verbose     # control echo on/off
    @cleanup = cleanup && cleanup.to_i
    @assets_definition = {} # load assets configuration

    # find the asset definition file:
    @definition_file = def_file || 'public/assets/definition.yml'
    echo 'use definition file: ', @definition_file

    parse_definitions
    set_setting_methods
    load_current_config
  end

  # clean previous config and update to new setting
  def update_config!
    @assets_config, @asset_file_list = nil, nil
    dir = File.dirname(output_file)
    FileUtils.mkdir_p(dir) unless File.directory?(dir)
    echo 'write development mapping as: ', development_mapping
    echo 'write production mapping as: ', production_mapping
    File.open(output_file, 'wb') do |fh|
      fh.write YAML.dump(
        'production' => production_mapping,
        'development' => development_mapping
      )
    end
  end

  def compile!
    @version = Time.now.gmtime.strftime('v%y%m%d%H%M')
    echo 'get current timestamp (utc) version number: ', @version

    # compile each asset definition based on asset config
    assets_config.each do |target, file_list|
      map_ary, compile_list = [], []

      # get the file name with version number
      new_file = production_path(target)

      # delete separate hosted assets from file list
      file_list.each { |f| url?(f) ? map_ary << f : compile_list << f }

      echo 'build compile list:', compile_list

      map_ary << compile_file!(new_file, compile_list) if compile_list.size > 0
      @production_mapping ||= {}
      @production_mapping[target] = map_ary
    end

    update_config!
    cleanup!
  end

  def cleanup!
    return unless @cleanup

    echo "cleanup asset files older than #{@cleanup} minutes"
    Dir[js_output_dir + '/*.js'].each do |f|
      next if used_filelist.include?(File.basename(f))
      if file_old?(f)
        echo 'delete old file: ', f
        FileUtils.rm f
      end
    end

    Dir[css_output_dir + '/*.css'].each do |f|
      next if used_filelist.include?(File.basename(f))
      if file_old?(f)
        echo 'delete old file: ', f
        FileUtils.rm f
      end
    end
  end

  def used_filelist
    used_filelist = []
    production_mapping.each do |t, flist|
      flist = [flist] unless flist.is_a?(Array)
      flist.each { |f| used_filelist << File.basename(f) unless url?(f) }
    end
    used_filelist
  end

  def file_old?(file)
    filestamp = file.match(/v(\d{10})/)[1]
    timeline = Time.now - 60 * @cleanup

    echo "compare #{filestamp} to #{timeline.gmtime.strftime('%y%m%d%H%M')}",
         file

    filestamp && timeline.gmtime.strftime('%y%m%d%H%M').to_i > filestamp.to_i
  end

  private

  def compile_file!(new_file, file_list)
    echo "compile #{new_file} from: ", file_list
    FileUtils.mkdir_p File.dirname(new_file)
    wh = File.open(new_file, 'wb')
    if get_type(new_file) == :css
      wh.write Sass::Engine.new(file_list.map { |f| File.read(f) }.join,
                                syntax: :scss, style: :compressed).render
    else
      echo 'perform closure compiler on file list: ', file_list
      wh.write Closure::Compiler.new.compile_files(file_list)
    end
    wh.close

    use_file = check_version(new_file)
    if use_file == new_file
      echo 'file updated ', use_file
    else
      echo 'file actually used for production: ', use_file
      FileUtils.rm new_file
    end

    production_file_url(use_file)
  end

  # map file to url (in development environment) for a single file
  def development_file_url(file)
    return file if url?(file)
    @settings['development_prefix'] + file.gsub(/^#{public_dir}\/?/, '')
  end

  def production_file_url(file)
    return file if url?(file)
    key = get_type(file) == :css ? 'css_url_prefix' : 'js_url_prefix'
    @settings[key] + File.basename(file)
  end

  # the resulted output mappings for development
  def development_mapping
    unless @development_mapping
      @development_mapping = {}
      assets_config.each do |target, files|
        @development_mapping[target] = files.map { |f| development_file_url f }
      end
    end
    @development_mapping
  end

  # the resulted output mappings for production
  def production_mapping
    @production_mapping || @current_config['production']
  end

  def parse_definitions
    YAML.load(File.read(@definition_file)).each do |target, file_list|
      if target == 'settings'
        update_settings(file_list)
      else
        fail "wrong file type #{target}" unless target.match(/\.(js|css)$/)
        fail "wrong file list for #{target}" unless file_list.is_a?(Array)
        @assets_definition[target] = file_list
      end
    end

    echo 'updated settings: ', @settings
    echo 'loaded definitions: ', @assets_definition
    self
  end

  # define directory full paths as accessors methods
  def set_setting_methods
    @settings.each do |k, v|
      next unless k.match(/_(dir|file)$/)
      echo 'defined method: ', k.to_s
      self.class.class_eval do
        var = v.size > 0 ? File.expand_path(v) : nil
        define_method(k.to_sym) { var }
      end
    end
  end

  def load_current_config
    if File.exists?(output_file)
      @current_config = YAML.load(File.read(output_file))
      echo 'loaded current config as: ', @current_config
    else
      @current_config = { production: {}, development: {} }
      echo 'not existing assets config file found'
    end
  end

  # file lists for all definitions
  def assets_config
    unless @assets_config
      @assets_config = {}
      @assets_definition.each do |target, file_list|
        type = get_type(target)
        @assets_config[target] = file_list.map do |file|
          url?(file) ? file : find_assets(file, type)
        end
      end

      echo 'find assets config from file definition: ', @assets_config
    end
    @assets_config
  end

  # # get all assets (with .css or .js extension) file under assets directory
  def asset_file_list
    unless @asset_file_list
      @asset_file_list = Dir[assets_dir + '/**/*.{css,js}'].to_a
      echo 'refreshed assets file list as: ', @asset_file_list
    end
    @asset_file_list
  end

  # map asset identifier to real file path,
  # type should be :js or :css
  def find_assets(ident, type)
    filefound, file_to_use = [], nil
    ident.gsub!(/\.#{type.to_s}$/, '')
    asset_file_list.each do |file|
      next unless file.match(/#{ident}/)
      next unless get_type(file) == type.to_sym
      filefound << file.gsub(assets_dir, '')
      file_to_use = file if !file_to_use || file.size < file_to_use.size
    end

    if filefound.size > 1
      echo "found more than one files for #{ident.blue} (#{type.to_s}):",
           filefound, true
      echo 'use: ', file_to_use.gsub(assets_dir, ''), true
    end
    fail "file for #{ident} (#{type.to_s}) not found" unless file_to_use
    file_to_use
  end

  # update settings from yaml config file, check setting value data type
  def update_settings(sets)
    sets.each do |k, v|
      fail "unknown setting '#{k}'" unless @settings.keys.include?(k)
      fail "setting '#{k}' must be a #{@settings[k].class.to_s}" unless
        v.is_a?(@settings[k].class)
      @settings[k] = v
    end
  end

  # check file with older versions, determine which file to use
  def check_version(file)
    filepattern = file.gsub(/\.v\d+\.(css|js)$/, "*.#{get_type(file).to_s}")
    Dir[filepattern].each do |f|
      next if f == file # skip the same file
      return f if File.read(f) == File.read(file)
    end
    file
  end

  def url?(file)
    file.match(%r!^(http:|https:)?//!)
  end

  # production file path for target file
  def production_path(target)
    type = get_type(target)
    dir = type == :css ? css_output_dir : js_output_dir
    dir + '/' + target.gsub(/\.(js|css)$/, ".#{@version}.#{type}")
  end

  def get_type(file)
    case
    when file.match(/.+\.css$/) then :css
    when file.match(/.+\.js$/) then :js
    else
      fail 'unsupported file type for ' +  file
    end
  end

  # echo message only if verbose specified to be true
  def echo(msg, var = nil, force = false)
    return unless @verbose || force
    msg += "\n" if var && var.inspect.size > 30

    print "[#{caller[0][/`.*'/][1..-2]}] ".green + msg

    if var
      print Term::ANSIColor.bold
      var.inspect.size > 30 ? PP.pp(var) : PP.singleline_pp(var)
      print Term::ANSIColor.reset
    end
    puts unless var && var.inspect.size > 30
  end
end
